-----------------------------------------------------------------------------
--
-- data
--   \damageDealt
--     \<damageDealerName>
--       -count, amount
--       \damageTypeDetails
--         \<damageTypeLocal>
--			 -count, amount
--       \attackDetails
--         \<attackName>
--            -count, amount, damageType
--            \hitDetails
--			    \<attackType> -- one of Hit, Critical Hit, Tick, Miss, Dodge, Resist, etc.
--                -count, amount
--   \damageTaken
--     \<damageTakerName>
--       -count, amount
--       \damageTypeDetails
--         \<damageTypeLocal>
--			 -count, amount
--       \attackDetails
--         \<attackName>
--            -count, amount, damageType
--            \hitDetails
--  			\<attackType>
--
-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentDamage")
L:RegisterTranslations("enUS", function() return {
	["Damage"] = true,
	["Damage Options."] = true,
	["Merge Pet Damage"] = true,
	["Merges damage dealt by pets with their owners data."] = true,

	["Damage Dealers"] = true,
	["Damage Takers"] = true,
	["Damage Dealt"] = true,
	["Fight Damage Dealt"] = true,
	["Damage Dealt Details"] = true,
	["Damage Dealt Hit Details"] = true,
	["DPS"] = true,

	["Damage Taken"] = true,
	["Damage Taken Details"] = true,
	["Damage Taken Hit Details"] = true,
	["DPS"] = true,
	["Fight Damage Taken"] = true,

	["Hit"] = true,
	["Critical Hit"] = true,
	["Tick"] = true,
	["Partial Absorb"] = true,
	["Partial Block"] = true,
	["Partial Resist"] = true,
	["[Melee]"] = true,
	
	["Pet: "] = true,

	["Damage Type"] = true,
	["Colors"] = true,
	["Color options"] = true,

	["Gather Damage Type Data"] = true,
	["Gather information regarding damage types taken."] = true,
	["Gather Attack Data"] = true,
	["Gather information about each attack used."] = true,
	["Gather Hit Data"] = true,
	["Gather hit information for each attack used."] = true,
} end)

-----------------------------------------------------------------------------

local AssessmentDamage = Assessment:NewModule("Damage", "AceEvent-2.0", "Parser-3.0")
AssessmentDamage.db = Assessment:AcquireDBNamespace("Damage")
Assessment:RegisterDefaults("Damage", "profile", {
	enable					= true,
	mergePetWithOwner		= true,
	gatherDamageTypeData	= true,
	gatherAttackData		= true,
	gatherHitData			= true,
	colors =
	{
		physical	= { 1, 1, 1 },
		holy		= { 1, 1, 0.5 },
		fire		= { 1, 0.3, 0 },
		nature		= { 0.5, 1, 0.2 },
		frost		= { 0.4, 0.6, 0.9 },
		shadow		= { 0.7, 0.4, 1 },
		arcane		= { 0.8, 0.8, 1 },
		unknown		= { 0, 0, 0 },
	}
})

local profile
local profileColors

-----------------------------------------------------------------------------

Assessment.TagData.damagedealt =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local damageDealt = fightData.damageDealt
				if not damageDealt then return 0 end
				
				if not damageDealt[data.key] then return 0 end
				return damageDealt[data.key].amount
			 end,
	data = "damage",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.damagetaken =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local damageTaken = fightData.damageTaken
				if not damageTaken then return 0 end
				
				if not damageTaken[data.key] then return 0 end
				return damageTaken[data.key].amount
			 end,
	data = "damage",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.absorbamount =
{
	tableData = "value.absorbAmount",
	numeric = true
}

Assessment.TagData.blockamount =
{
	tableData = "value.blockAmount",
	numeric = true
}

Assessment.TagData.misscount =
{
	tableData = "value.count",
	numeric = true
}

Assessment.TagData.resistamount =
{
	tableData = "value.resistAmount",
	numeric = true
}

-----------------------------------------------------------------------------

Assessment.ListType.DamageDealers =
{
	method = function(fightData)
				if not fightData or not fightData.damageDealt then return function() return nil end end
				return pairs(fightData.damageDealt)
			 end,
	data = "damage",
	description = L["Damage Dealers"],
}

Assessment.ListType.DamageTakers =
{
	method = function(fightData)
				if not fightData or not fightData.damageTaken then return function() return nil end end
				return pairs(fightData.damageTaken)
			 end,
	data = "damage",
	description = L["Damage Takers"],
}

-----------------------------------------------------------------------------

Assessment.DefaultDisplaySets.DamageDealtHitDetails =
{
	titleTag			= "[windowDisplaySetParameter(2)]: [windowDisplaySetParameter(4)]",
	listType			= "EnumerateTable", 
	barValueTag			= "amount",
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { tag = "[rank]. [key] ([count])" }, { tag = "[barValue:short]" } },
	description			= L["Damage Dealt Hit Details"],
	selectable			= false,
--	colorTypeList		= { "Spell" },
}

Assessment.DefaultDisplaySets.DamageDealtDetails =
{
	titleTag			= "[(fightIndex == 1):\"Fight \"]Damage Dealt: [windowDisplaySetParameter(2)]",
	listType			= "EnumerateTable", 
	barValueTag			= "amount",
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { tag = "[rank]. [key|\"White Damage\"] ([count])" }, { tag = "[barValue:short]" } },
	description			= L["Damage Dealt Details"],
	selectable			= false,
	onClickType			= "PUSH_DISPLAY_SET",
	onClickDisplaySet	= L["Damage Dealt Hit Details"],
	onClickData			= "damageDealt,[windowDisplaySetParameter(2)],attackDetails,[key],hitDetails",
	colorTypeList		= { "SpellDamageType" },
}

Assessment.DefaultDisplaySets.DamageDealt =
{
	titleTag			= "Damage Dealt: [totalBarValue:short]",
	listType			= "GroupMembers", 
	barValueTag			= "damageDealt", 
	filterTag			= "barValue ~= 0",
	fightIndex			= 0,
	text				= { { }, { tag= "[barValue:short]" } },
	description			= L["Damage Dealt"],
	onClickType			= "PUSH_DISPLAY_SET",
	onClickDisplaySet	= L["Damage Dealt Details"],
	onClickData			= "damageDealt,[key],attackDetails",
	segmentEnable		= false,
	segmentDisplaySet	= L["Damage Type"],
	segmentData			= "damageDealt,[key],damageTypeDetails",
}

Assessment.DefaultDisplaySets.FightDamageDealt =
{
	titleTag			= "Fight Damage Dealt: [totalBarValue:short]",
	listType			= "GroupMembers", 
	barValueTag			= "damageDealt", 
	filterTag			= "barValue ~= 0",
	fightIndex			= 1,
	text				= { { }, { tag= "[barValue:short]" } },
	description			= L["Fight Damage Dealt"],
	onClickType			= "PUSH_DISPLAY_SET",
	onClickDisplaySet	= L["Damage Dealt Details"],
	onClickData			= "damageDealt,[key],attackDetails",
}

Assessment.DefaultDisplaySets.DPS =
{
	titleTag			= "DPS: [combatTime:formatDuration]",
	listType			= "GroupMembers", 
	barValueTag			= "damageDealt / maximum(combatTime, 1)", 
	filterTag			= "barValue ~= 0",
	fightIndex			= 1,
	text				= { { }, { tag= "[barValue:format(\"%.1f\")]" } },
	description			= L["DPS"],
	onClickType			= "PUSH_DISPLAY_SET",
	onClickDisplaySet	= L["Damage Dealt Details"],
	onClickData			= "damageDealt,[key],attackDetails",
	favorite			= false,
	segmentEnable		= false,
	segmentDisplaySet	= L["Damage Type"],
	segmentData			= "damageDealt,[key],damageTypeDetails",
}

Assessment.DefaultDisplaySets.DamageTakentHitDetails =
{
	titleTag			= "[windowDisplaySetParameter(2)]: [windowDisplaySetParameter(4)]",
	listType			= "EnumerateTable", 
	barValueTag			= "amount",
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { tag = "[rank]. [key] ([count])" }, { tag = "[barValue:short]" } },
	description			= L["Damage Taken Hit Details"],
	selectable			= false,
--	colorTypeList		= { "Spell" },
}

Assessment.DefaultDisplaySets.DamageTakenDetails =
{
	titleTag			= "[(fightIndex == 1):\"Fight \"]Damage Taken: [windowDisplaySetParameter(2)]",
	listType			= "EnumerateTable", 
	barValueTag			= "amount",
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { tag = "[rank]. [key|\"White Damage\"] ([count])" }, { tag = "[barValue:short]" } },
	description			= L["Damage Taken Details"],
	selectable			= false,
	onClickType			= "PUSH_DISPLAY_SET",
	onClickDisplaySet	= L["Damage Taken Hit Details"],
	onClickData			= "damageTaken,[windowDisplaySetParameter(2)],attackDetails,[key],hitDetails",
	colorTypeList		= { "SpellDamageType" },
}

Assessment.DefaultDisplaySets.DamageType =
{
	titleTag			= "[windowDisplaySetParameter(1)]",
	listType			= "EnumerateTable", 
	barValueTag			= "amount", 
	filterTag			= "1",
	sortKeyTag			= "barValue",
	fightIndex			= -1,
	text				= { { }, { tag= "[barValue:short]" } },
	description			= L["Damage Type"],
	colorTypeList		= { "DamageType", count = 1 },
	favorite			= false,
	selectable			= false,
}

Assessment.DefaultDisplaySets.DamageTaken =
{
	titleTag			= "Damage Taken: [totalBarValue:short]",
	listType			= "GroupMembers", 
	barValueTag			= "damageTaken", 
	filterTag			= "barValue ~= 0",
	fightIndex			= 0,
	text				= { { }, { tag= "[barValue:short]" } },
	description			= L["Damage Taken"],
	onClickType			= "PUSH_DISPLAY_SET",
	onClickDisplaySet	= L["Damage Taken Details"],
	onClickData			= "damageTaken,[key],attackDetails",
	segmentEnable		= true,
	segmentDisplaySet	= L["Damage Type"],
	segmentData			= "damageTaken,[key],damageTypeDetails",
}

Assessment.DefaultDisplaySets.FightDamageTaken =
{
	titleTag			= "Fight Damage Taken: [totalBarValue:short]",
	listType			= "GroupMembers", 
	barValueTag			= "damageTaken", 
	filterTag			= "barValue ~= 0",
	fightIndex			= 1,
	text				= { { }, { tag= "[barValue:short]" } },
	description			= L["Fight Damage Taken"],
	onClickType			= "PUSH_DISPLAY_SET",
	onClickDisplaySet	= L["Damage Taken Details"],
	onClickData			= "damageTaken,[key],attackDetails",
}

-----------------------------------------------------------------------------

local function getColor(passValue)
	return unpack(profileColors[passValue])
end

local function setColor(passValue,r, g, b)
	profileColors[passValue] = { r, g, b }
	Assessment:RedrawAll()
end

Assessment.Options.args.data.args.damage =
{
	type = "group",
	name = L["Damage"],
	desc = L["Damage Options."],
	args =
	{
		gatherDamageType =
		{
			type = "toggle",
			name = L["Gather Damage Type Data"],
			desc = L["Gather information regarding damage types taken."],
			get = function() return profile.gatherDamageTypeData end,
			set = function(v) profile.gatherDamageTypeData = v end,
		},
		gatherAttackData =
		{
			type = "toggle",
			name = L["Gather Attack Data"],
			desc = L["Gather information about each attack used."],
			get = function() return profile.gatherAttackData end,
			set = function(v) profile.gatherAttackData = v end,
		},
		gatherHitData =
		{
			type = "toggle",
			name = L["Gather Hit Data"],
			desc = L["Gather hit information for each attack used."],
			disabled = function() return not profile.gatherAttackData end,
			get = function() return profile.gatherHitData end,
			set = function(v) profile.gatherHitData = v end,
		},
		mergePetWithOwner =
		{
			type = "toggle",
			name = L["Merge Pet Damage"],
			desc = L["Merges damage dealt by pets with their owners data."],
			get = function() return profile.mergePetWithOwner end,
			set = function(v) profile.mergePetWithOwner = v end,
		},
		color =
		{
			type = "group",
			name = L["Colors"],
			desc = L["Color options"],
			args =
			{
				-- Filled in below
			},
		},
	}
}

-----------------------------------------------------------------------------

local localDamageTypeToColorName = 
{ 
	[SPELL_SCHOOL0_CAP] = "physical",
	[SPELL_SCHOOL1_CAP] = "holy",
	[SPELL_SCHOOL2_CAP] = "fire",
	[SPELL_SCHOOL3_CAP] = "nature",
	[SPELL_SCHOOL4_CAP] = "frost",
	[SPELL_SCHOOL5_CAP] = "shadow",
	[SPELL_SCHOOL6_CAP] = "arcane",
	[UNKNOWN]			= "unknown"
}

for k,v in pairs(localDamageTypeToColorName) do
	Assessment.Options.args.data.args.damage.args.color.args[v] =
	{
		type = "color",
		name = k,
		desc = k,
		get = getColor,
		set = setColor,
		passValue = v
	}
end

-----------------------------------------------------------------------------

Assessment.ColorType.DamageType = 
{
	method = function(fightData, data)
				local colorName = localDamageTypeToColorName[data.key]
				if not colorName then return nil end
				return profileColors[colorName]
			 end,
	description = L["Damage Type"]
}

Assessment.ColorType.SpellDamageType = 
{
	method = function(fightData, data)
				local damageType = data.value.damageType
				if not damageType then return end
				
				return profileColors[damageType]
			 end,
	description = L["Damage Type"]
}

-----------------------------------------------------------------------------

Assessment.DefaultTags.DPS =
{
	tag			= "[(damageDealt / maximum(combatTime, 1)):format(\"%.1f\")]",
	fightIndex	= 1,
	description	= L["DPS"],
}

-----------------------------------------------------------------------------

function AssessmentDamage:OnInitialize()
	profile = self.db.profile
	profileColors = profile.colors
end

function AssessmentDamage:OnEnable()
	self:RegisterParserEvent({ eventType = "Damage" }, "OnParserDamageEvent")
	self:RegisterParserEvent({ eventType = "Miss" }, "OnParserMissEvent")
end

function AssessmentDamage:OnDisable()
end

-----------------------------------------------------------------------------

local function AccumulateData(data, eventData)
	data.count = data.count + 1
	data.amount = data.amount + eventData.amount
end

local petToOwnerTable = { pet = "player" }
for i=1,4 do petToOwnerTable["partypet"..i] = "party"..i end
for i=1,40 do petToOwnerTable["raidpet"..i] = "raid"..i end

-----------------------------------------------------------------------------

function AssessmentDamage:NewDataTable()
	local data = self:NewTable()
	data.count = 0
	data.amount = 0
	return data
end

-----------------------------------------------------------------------------

function AssessmentDamage:UpdateDamageData(dataTable, member, eventData, attackType, abilityName)
	local data = dataTable[member]
	if not data then
		data = self:NewDataTable()
		data.attackDetails = self:NewTable()
		data.damageTypeDetails = self:NewTable()
		dataTable[member] = data
	end
	AccumulateData(data, eventData)

	if profile.gatherDamageTypeData then
		local damageTypeDetailsParent = data.damageTypeDetails
		local damageTypeDetails = damageTypeDetailsParent[eventData.damageTypeLocal]
		if not damageTypeDetails then
			damageTypeDetails = self:NewDataTable()
			damageTypeDetailsParent[eventData.damageTypeLocal] = damageTypeDetails
		end
		AccumulateData(damageTypeDetails, eventData)
	end

	if profile.gatherAttackData then	
		local attackDetailsParent = data.attackDetails
		local attackDetails = attackDetailsParent[abilityName]
		if not attackDetails then
			attackDetails = self:NewDataTable()
			attackDetails.hitDetails = self:NewTable()
			attackDetailsParent[abilityName] = attackDetails
		end
		attackDetails.damageType = localDamageTypeToColorName[eventData.damageTypeLocal]
		AccumulateData(attackDetails, eventData)
		
		if profile.gatherHitData then
			local hitDetails = attackDetails.hitDetails
			local attackTypeDetails = hitDetails[attackType]
			if not attackTypeDetails then
				attackTypeDetails = self:NewDataTable()
				hitDetails[attackType] = attackTypeDetails
			end
			AccumulateData(attackTypeDetails, eventData)
			
			if eventData.absorbAmount ~= 0 then
				local type = L["Partial Absorb"]
				local attackTypeDetails = hitDetails[type]
				if not attackTypeDetails then
					attackTypeDetails = self:NewDataTable()
					hitDetails[type] = attackTypeDetails
				end
				attackTypeDetails.count = attackTypeDetails.count + 1
				attackTypeDetails.amount = attackTypeDetails.amount + eventData.absorbAmount
			end

			if eventData.blockAmount ~= 0 then
				local type = L["Partial Block"]
				local attackTypeDetails = hitDetails[type]
				if not attackTypeDetails then
					attackTypeDetails = self:NewDataTable()
					hitDetails[type] = attackTypeDetails
				end
				attackTypeDetails.count = attackTypeDetails.count + 1
				attackTypeDetails.amount = attackTypeDetails.amount + eventData.blockAmount
			end

			if eventData.resistAmount ~= 0 then
				local type = L["Partial Resist"]
				local attackTypeDetails = hitDetails[type]
				if not attackTypeDetails then
					attackTypeDetails = self:NewDataTable()
					hitDetails[type] = attackTypeDetails
				end
				attackTypeDetails.count = attackTypeDetails.count + 1
				attackTypeDetails.amount = attackTypeDetails.amount + eventData.resistAmount
			end
		end
	end
end

-----------------------------------------------------------------------------

function AssessmentDamage:OnParserDamageEvent(eventData)
	local attackType
	if eventData.isDoT then
		attackType = L["Tick"]
	elseif eventData.isCrit then
		attackType = L["Critical Hit"]
	else
		attackType = L["Hit"]
	end
	
	local abilityName = eventData.abilityName or L["[Melee]"]

	local sourceName = eventData.sourceName
	if profile.mergePetWithOwner then
		local sourceOwnerID = petToOwnerTable[eventData.sourceID]
		if sourceOwnerID then
			abilityName = L["Pet: "]..abilityName
			sourceName = UnitName(sourceOwnerID)
		end
	end
	
	for data in self:IterateUpdateDataTables() do
		data.dataUpdateTimes.damage = GetTime()
		
		local damageDealt = data.damageDealt
		if not damageDealt then
			damageDealt = self:NewTable()
			data.damageDealt = damageDealt
		end

		local damageTaken = data.damageTaken
		if not damageTaken then
			damageTaken = self:NewTable()
			data.damageTaken = damageTaken
		end

		self:UpdateDamageData(damageDealt, sourceName, eventData, attackType, abilityName)
		self:UpdateDamageData(damageTaken, eventData.recipientName, eventData, attackType, abilityName)
	end
end

-----------------------------------------------------------------------------

function AssessmentDamage:UpdateMissData(dataTable, member, eventData, abilityName)
	local data = dataTable[member]
	if not data then
		data = self:NewDataTable()
		data.attackDetails = self:NewTable()
		data.damageTypeDetails = self:NewTable()
		dataTable[member] = data
	end

	if profile.gatherAttackData then
		local attackDetailsParent = data.attackDetails
		local attackDetails = attackDetailsParent[abilityName]
		if not attackDetails then
			attackDetails = self:NewDataTable()
			attackDetails.hitDetails = self:NewTable()
			attackDetailsParent[abilityName] = attackDetails
		end
		
		if profile.gatherHitData then		
			local hitDetails = attackDetails.hitDetails
			local attackTypeDetails = hitDetails[eventData.missTypeLocal]
			if not attackTypeDetails then
				attackTypeDetails = self:NewDataTable()
				hitDetails[eventData.missTypeLocal] = attackTypeDetails
			end
			attackTypeDetails.count = attackTypeDetails.count + 1
		end
	end
end

-----------------------------------------------------------------------------

function AssessmentDamage:OnParserMissEvent(eventData)
	local abilityName = eventData.abilityName or L["[Melee]"]
	
	local sourceName = eventData.sourceName
	if profile.mergePetWithOwner then
		local sourceOwnerID = petToOwnerTable[eventData.sourceID]
		if sourceOwnerID then
			abilityName = L["Pet: "]..abilityName
			sourceName = UnitName(sourceOwnerID)
		end
	end
	
	for data in self:IterateUpdateDataTables() do
		data.dataUpdateTimes.damage = GetTime()
		
		local damageDealt = data.damageDealt
		if not damageDealt then
			damageDealt = self:NewTable()
			data.damageDealt = damageDealt
		end

		local damageTaken = data.damageTaken
		if not damageTaken then
			damageTaken = self:NewTable()
			data.damageTaken = damageTaken
		end

		self:UpdateMissData(damageDealt, sourceName, eventData, abilityName)
		self:UpdateMissData(damageTaken, eventData.recipientName, eventData, abilityName)
	end
end

-----------------------------------------------------------------------------
